use ureq;
use serde_json;
use rayon::prelude::*;

fn main() {
    // Grab minecraft version manifest
    let response = ureq::get("https://launchermeta.mojang.com/mc/game/version_manifest.json").call();
    let version_manifest = response.unwrap().into_string().unwrap();
    // Get the list of all versions
    let version_manifest: serde_json::Value = serde_json::from_str(&version_manifest).unwrap();
    let version_objs = version_manifest["versions"].as_array().unwrap();
    let _version_ids: Vec<String> = version_objs.iter().map(|v| v["id"].as_str().unwrap().to_string()).collect();
    // Grab version 1.20.1
    let latest = version_objs.iter().find(|v| v["id"].as_str().unwrap() == "1.15").unwrap();
    // Grab the version's json
    let response = ureq::get(latest["url"].as_str().unwrap()).call();
    // Version URL
    println!("Version URL: {}", latest["url"].as_str().unwrap());
    let version_json = response.unwrap().into_string().unwrap();
    let version_json: serde_json::Value = serde_json::from_str(&version_json).unwrap();
    println!("Latest version jar: {}", version_json["downloads"]["client"]["url"].as_str().unwrap());
    // Grab the list of library URLs
    let libraries = version_json["libraries"].as_array().unwrap();
    let mut library_urls: Vec<String> = Vec::new();
    for l in libraries {
        // Check the rules for linux
        if l["rules"].is_array() {
            let rules = l["rules"].as_array().unwrap();
            if rules.iter().any(|r| r["action"].as_str().unwrap_or("deny") == "allow" && r["os"]["name"].as_str().unwrap_or("linux") != "osx"){
                let url = l["downloads"]["artifact"]["url"].as_str();
                if url.is_some() {
                    library_urls.push(String::from(url.unwrap()));
                } else {
                    println!("Artifact URL not found: {:?}", l);
                }
            }
        } else {
            let url = l["downloads"]["artifact"]["url"].as_str();
            if url.is_some() {
                library_urls.push(String::from(url.unwrap()));
            } else {
                println!("Artifact URL not found: {:?}", l);
            }
        }
    }
    let native_urls: Vec<String> = libraries.iter().filter_map(|l| {
        if l["downloads"]["classifiers"].is_object() {
            let classifiers = l["downloads"]["classifiers"].as_object().unwrap();
            if classifiers.contains_key("natives-linux") {
                return Some(String::from(classifiers["natives-linux"]["url"].as_str().unwrap()));
            }
            // if classifiers.contains_key("natives-windows") {
            //     return Some(String::from(classifiers["natives-windows"]["url"].as_str().unwrap()));
            // }
            // if classifiers.contains_key("natives-osx") {
            //     return Some(String::from(classifiers["natives-osx"]["url"].as_str().unwrap()));
            // }
        }
        None
    }).collect();
    println!("Library URLs: {:?}", library_urls);
    // Grab the asset index
    let asset_index = version_json["assetIndex"].as_object().unwrap();
    println!("Asset index url: {}", asset_index["url"].as_str().unwrap());
    // Grab the asset index json
    let response = ureq::get(asset_index["url"].as_str().unwrap()).call();
    let asset_index_json = response.unwrap().into_string().unwrap();
    let asset_index_json: serde_json::Value = serde_json::from_str(&asset_index_json).unwrap();

    // Save the asset index into indexes
    std::fs::create_dir_all("assets/indexes").unwrap();
    std::fs::write(format!("assets/indexes/{}.json", asset_index["id"].as_str().unwrap()), asset_index_json.to_string()).unwrap();

    // Download all the assets in parallel
    let objects = asset_index_json["objects"].as_object().unwrap().to_owned();
    let objects_vec: Vec<serde_json::Value> = objects.iter().map(|(_key, value)| value.clone()).collect();
    std::fs::create_dir_all("assets/objects/").unwrap();
    let index: std::sync::Mutex<i32> = std::sync::Mutex::new(0);
    // Wait for all downloads to finish
    objects_vec.par_iter().map(|value| {
        // Print progress in percent
        println!("{}%", ((*index.lock().unwrap() as f32 / objects.len() as f32) * 100.0) as i32);
        let hash = value["hash"].as_str().unwrap();
        let hash_prefix = &hash[0..2];
        // Skip if the asset already exists
        *index.lock().unwrap() += 1;
        if std::fs::metadata(format!("assets/objects/{}/{}", hash_prefix, hash)).is_ok() {
            return;
        }
        let asset_url = format!("https://resources.download.minecraft.net/{}/{}", hash_prefix, hash);
        println!("Downloading asset: {}", asset_url);
        let response = ureq::get(&asset_url).call();
        let response = response.unwrap();
        let len: usize = response.header("Content-Length")
            .unwrap()
            .parse().unwrap();
        let mut bytes: Vec<u8> = Vec::with_capacity(len);
        response.into_reader()
            .read_to_end(&mut bytes).unwrap();
        // Make hash prefix dir
        std::fs::create_dir_all(format!("assets/objects/{}", hash_prefix)).unwrap();
        // Write the asset
        std::fs::write(format!("assets/objects/{}/{}", hash_prefix, hash), bytes).unwrap();
    }).collect::<Vec<_>>();
    // Download the client jar
    let response = 
    ureq::get(version_json["downloads"]["client"]["url"].as_str().unwrap()).call();
    let response = response.unwrap();
    let len: usize = response.header("Content-Length")
        .unwrap()
        .parse().unwrap();
    let mut bytes: Vec<u8> = Vec::with_capacity(len);
    response.into_reader()
        .read_to_end(&mut bytes).unwrap();
    std::fs::create_dir_all("versions").unwrap();
    std::fs::write(format!("versions/{}.jar", latest["id"].as_str().unwrap()), bytes).unwrap();

    // Download the libraries
    println!("Downloading libraries");
    let mut index: i32 = 0;
    let library_len = library_urls.len();
    for library_url in &library_urls {
        println!("{}%", ((index as f32 / library_len as f32) * 100.0) as i32);
        // Skip if the library already exists
        index += 1;
        if std::fs::metadata(format!("libraries/{}", library_url.split("/").last().unwrap())).is_ok() {
            continue;
        }
        let response = ureq::get(&library_url).call();
        let response = response.unwrap();
        let len: usize = response.header("Content-Length")
            .unwrap()
            .parse().unwrap();
        let mut bytes: Vec<u8> = Vec::with_capacity(len);
        response.into_reader()
            .read_to_end(&mut bytes).unwrap();
        let path = format!("libraries/{}", library_url.split("/").last().unwrap());
        std::fs::create_dir_all("libraries").unwrap();
        std::fs::write(path, bytes).unwrap();
    }

    // Download the natives
    println!("Downloading natives");
    std::fs::create_dir_all(format!("natives/{}", latest["id"].as_str().unwrap()).as_str()).unwrap();
    let mut index: i32 = 0;
    let native_len = native_urls.len();
    for native_url in &native_urls {
        println!("{}%", ((index as f32 / native_len as f32) * 100.0) as i32);
        // Skip if the native already exists
        index += 1;
        if std::fs::metadata(format!("natives/{}/{}", latest["id"].as_str().unwrap(), native_url.split("/").last().unwrap())).is_ok() {
            continue;
        }
        let response = ureq::get(&native_url).call();
        let response = response.unwrap();
        let len: usize = response.header("Content-Length")
            .unwrap()
            .parse().unwrap();
        let mut bytes: Vec<u8> = Vec::with_capacity(len);
        response.into_reader()
            .read_to_end(&mut bytes).unwrap();
        let path = format!("natives/{}/{}", latest["id"].as_str().unwrap(), native_url.split("/").last().unwrap());
        std::fs::write(path, bytes).unwrap();
    }

    // Unzip the natives with unzip command
    let files_in_natives = std::fs::read_dir(format!("natives/{}", latest["id"].as_str().unwrap()).as_str()).unwrap();
    // Filter for .jar
    for file in files_in_natives {
        let file = file.unwrap();
        let path = file.path();
        let path = path.to_str().unwrap();
        if path.ends_with(".jar") {
            std::process::Command::new("unzip")
                .current_dir(format!("natives/{}", latest["id"].as_str().unwrap()).as_str())
                .arg("-o")
                .arg(file.file_name())
                .status()
                .unwrap();
        }
    }

    // Figure out Java version from MC version
    let mc_version = latest["id"].as_str().unwrap();
    let mc_minor = mc_version.split(".").nth(1).unwrap().parse::<i32>().unwrap();
    let java_ver = match mc_minor {
        ..=12 => "8",
        ..=16 => "11",
        ..=19 => "17",
        20.. => "21",
        _ => "8"
    };


    // Build the launch command
    let mut launch_command = String::from(format!("/usr/lib/jvm/java-{}-openjdk/bin/java -cp ", java_ver));

    // Build the classpath
    let mut classpath = String::new();
    for library_url in &library_urls {
        let path = format!("libraries/{}", library_url.split("/").last().unwrap());
        classpath.push_str(&path);
        classpath.push_str(":");
    }
    classpath.push_str(format!("versions/{}.jar", latest["id"].as_str().unwrap()).as_str());

    launch_command.push_str(&classpath);

    // Add the natives, absolute path
    launch_command.push_str(format!(" -Djava.library.path=$(pwd)/natives/{}", latest["id"].as_str().unwrap()).as_str());

    // Append the Xmx and Xms
    launch_command.push_str(" -Xmx2G -Xms2G");

    // Append the main class
    launch_command.push_str(format!(" net.minecraft.client.main.Main --username {} --version {} --gameDir .", "Dev", latest["id"].as_str().unwrap()).as_str());

    // Append the assetDir
    launch_command.push_str(" --assetsDir assets");

    // Append the assetIndex
    launch_command.push_str(format!(" --assetIndex {}", asset_index["id"].as_str().unwrap()).as_str());

    // Append the uuid
    launch_command.push_str(" --uuid 8d3f4f3c-0a1c-4b1d-8f1d-2f4b3f4b3f4b");

    // Append the bogus access token
    launch_command.push_str(" --accessToken missingno");

    // Append empty userProperties
    launch_command.push_str(" --userProperties {}");

    // Run the command
    println!("Running command: {}", launch_command);
    std::process::Command::new("sh")
        .arg("-c")
        .arg(launch_command)
        .status()
        .unwrap();

}
