use serde::{Deserialize, Serialize};
use serde_with::skip_serializing_none;
use std::collections::HashMap;

#[derive(Serialize, Deserialize, Debug)]
pub struct VersionManifest {
    pub latest: Latest,
    pub versions: Vec<Version>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Latest {
    pub release: String,
    pub snapshot: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Version {
    pub id: String,
    #[serde(rename = "type")]
    pub type_: String,
    pub url: String,
    pub time: String,
    #[serde(rename = "releaseTime")]
    pub release_time: String,
}

#[skip_serializing_none]
#[derive(Serialize, Deserialize, Debug)]
pub struct AssetIndexUrl {
    pub id: String,
    pub sha1: String,
    pub size: u64,
    pub url: String,
}

#[skip_serializing_none]
#[derive(Serialize, Deserialize, Debug)]
pub struct VersionJson {
    pub arguments: Option<Arguments>,
    #[serde(rename = "assetIndex")]
    pub asset_index: AssetIndexUrl,
    pub assets: String,
    #[serde(rename = "complianceLevel")]
    pub compliance_level: Option<u32>,
    pub downloads: Downloads,
    pub id: String,
    #[serde(rename = "javaVersion")]
    pub java_version: Option<JavaVersion>,
    pub libraries: Vec<Library>,
    pub logging: Option<Logging>,
    #[serde(rename = "mainClass")]
    pub main_class: String,
    #[serde(rename = "minimumLauncherVersion")]
    pub minimum_launcher_version: u32,
    #[serde(rename = "releaseTime")]
    pub release_time: String,
    pub time: String,
    #[serde(rename = "type")]
    pub type_: String,
    #[serde(rename = "minecraftArguments")]
    pub minecraft_arguments: Option<String>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Arguments {
    pub game: Option<Vec<GameArgument>>,
    pub jvm: Option<Vec<JVMArgument>>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(untagged)]
pub enum GameArgument {
    String(String),
    Complex(GameArgumentComplex),
}

#[derive(Serialize, Deserialize, Debug)]
pub struct GameArgumentComplex {
    pub rules: Option<Vec<Rule>>,
    pub value: ValueType,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(untagged)]
pub enum JVMArgument {
    String(String),
    Complex(JVMArgumentComplex),
}

#[derive(Serialize, Deserialize, Debug)]
pub struct JVMArgumentComplex {
    pub rules: Option<Vec<Rule>>,
    pub value: ValueType,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(untagged)]
pub enum ValueType {
    String(String),
    VecString(Vec<String>),
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Rule {
    pub action: String,
    pub os: Option<OS>,
    pub features: Option<HashMap<String, bool>>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct OS {
    pub arch: Option<String>,
    pub name: Option<String>,
    pub version: Option<String>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct AssetIndex {
    pub objects: HashMap<String, AssetObject>,
    #[serde(rename = "virtual")]
    pub virtual_: Option<bool>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct AssetObject {
    pub hash: String,
    pub size: u64,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Downloads {
    pub client: DownloadFile,
    #[serde(rename = "client_mappings")]
    pub client_mappings: Option<DownloadFile>,
    pub server: DownloadFile,
    #[serde(rename = "server_mappings")]
    pub server_mappings: Option<DownloadFile>,
    #[serde(rename = "windows_server")]
    pub windows_server: Option<DownloadFile>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct DownloadFile {
    pub sha1: String,
    pub size: u64,
    pub url: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct JavaVersion {
    pub component: String,
    #[serde(rename = "majorVersion")]
    pub major_version: u32,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Library {
    pub downloads: LibraryDownloads,
    pub name: String,
    pub rules: Option<Vec<Rule>>,
    pub extract: Option<Extract>,
    pub natives: Option<Natives>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(untagged)]
pub enum LibraryDownloads {
    Artifact(LibraryArtifact),
    Classifiers(LibraryClassifiers),
}

#[derive(Serialize, Deserialize, Debug)]
pub struct LibraryArtifact {
    pub artifact: LibraryFile,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct LibraryClassifiers {
    pub classifiers: Classifiers,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Classifiers {
    #[serde(rename = "natives-linux")]
    pub natives_linux: Option<LibraryFile>,
    #[serde(rename = "natives-osx")]
    pub natives_osx: Option<LibraryFile>,
    #[serde(rename = "natives-windows")]
    pub natives_windows: Option<LibraryFile>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct LibraryFile {
    pub path: String,
    pub sha1: String,
    pub size: u64,
    pub url: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Extract {
    pub exclude: Vec<String>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Natives {
    pub linux: String,
    pub osx: String,
    pub windows: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct Logging {
    pub client: LoggingClient,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct LoggingClient {
    pub argument: String,
    pub file: LoggingFile,
    #[serde(rename = "type")]
    pub type_: String,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct LoggingFile {
    pub id: String,
    pub sha1: String,
    pub size: u64,
    pub url: String,
}
